@@ -14,7 +14,7 @@ use overload
 
 use if ($] >= 5.010), overload => '~~'  => "matches";
 
-our $VERSION = '1.11_01';
+our $VERSION = '1.999_01';
 
 my $PACKAGE = __PACKAGE__;  # Useful to have a scalar for hash keys.
 
@@ -39,9 +39,6 @@ autodie::exception - Exceptions from autodying functions.
 
 =head1 DESCRIPTION
 
-B<NOTE!  This is BETA code.  It is NOT the final release.  Implementation
-and interface may change!>
-
 When an L<autodie> enabled function fails, it generates an
 C<autodie::exception> object.  This can be interrogated to
 determine further information about the error that occurred.
@@ -174,7 +171,9 @@ C<CORE::> is assumed.
 
 For a string that does start with a colon, if the subroutine
 throwing the exception I<does> that behaviour.  For example, the
-C<CORE::open> subroutine does C<:file>, C<:io>, and C<:CORE>.
+C<CORE::open> subroutine does C<:file>, C<:io> and C<:all>.
+
+See L<autodie/CATEGORIES> for futher information.
 
 =back
 
@@ -240,10 +239,93 @@ work closely with the C<autodie::exception> model.
 #        get used in most programs.
 
 my %formatter_of = (
-    'CORE::close' => \&_format_close,
-    'CORE::open'  => \&_format_open,
+    'CORE::close'   => \&_format_close,
+    'CORE::open'    => \&_format_open,
+    'CORE::dbmopen' => \&_format_dbmopen,
+    'CORE::flock'   => \&_format_flock,
 );
 
+# TODO: Our tests only check LOCK_EX | LOCK_NB is properly
+# formatted.  Try other combinations and ensure they work
+# correctly.
+
+sub _format_flock {
+    my ($this) = @_;
+
+    require Fcntl;
+
+    my $filehandle = $this->args->[0];
+    my $raw_mode   = $this->args->[1];
+
+    my $mode_type;
+    my $lock_unlock;
+
+    if ($raw_mode & Fcntl::LOCK_EX() ) {
+        $lock_unlock = "lock";
+        $mode_type = "for exclusive access";
+    }
+    elsif ($raw_mode & Fcntl::LOCK_SH() ) {
+        $lock_unlock = "lock";
+        $mode_type = "for shared access";
+    }
+    elsif ($raw_mode & Fcntl::LOCK_UN() ) {
+        $lock_unlock = "unlock";
+        $mode_type = "";
+    }
+    else {
+        # I've got no idea what they're trying to do.
+        $lock_unlock = "lock";
+        $mode_type = "with mode $raw_mode";
+    }
+
+    my $cooked_filehandle;
+
+    if ($filehandle and not ref $filehandle) {
+
+        # A package filehandle with a name!
+
+        $cooked_filehandle = " $filehandle";
+    }
+    else {
+        # Otherwise we have a scalar filehandle.
+
+        $cooked_filehandle = '';
+
+    }
+
+    local $! = $this->errno;
+
+    return "Can't $lock_unlock filehandle$cooked_filehandle $mode_type: $!";
+
+}
+
+# Default formatter for CORE::dbmopen
+sub _format_dbmopen {
+    my ($this) = @_;
+    my @args   = @{$this->args};
+
+    # TODO: Presently, $args flattens out the (usually empty) hash
+    # which is passed as the first argument to dbmopen.  This is
+    # a bug in our args handling code (taking a reference to it would
+    # be better), but for the moment we'll just examine the end of
+    # our arguments list for message formatting.
+
+    my $mode = $args[-1];
+    my $file = $args[-2];
+
+    # If we have a mask, then display it in octal, not decimal.
+    # We don't do this if it already looks octalish, or doesn't
+    # look like a number.
+
+    if ($mode =~ /^[^\D0]\d+$/) {
+        $mode = sprintf("0%lo", $mode);
+    };
+
+    local $! = $this->errno;
+
+    return "Can't dbmopen(%hash, '$file', $mode): '$!'";
+}
+
 # Default formatter for CORE::close
 
 sub _format_close {
@@ -257,33 +339,81 @@ sub _format_close {
         return "Can't close filehandle '$close_arg': '$!'";
     }
 
+    # TODO - This will probably produce an ugly error.  Test and fix.
     return "Can't close($close_arg) filehandle: '$!'";
 
 }
 
 # Default formatter for CORE::open
-# Currently only works with 3-arg open.
-# TODO: Pretty printing for 2-arg (and 1-arg?) open.
+
+use constant _FORMAT_OPEN => "Can't open '%s' for %s: '%s'";
+
+sub _format_open_with_mode {
+    my ($this, $mode, $file, $error) = @_;
+
+    my $wordy_mode;
+
+    if    ($mode eq '<')  { $wordy_mode = 'reading';   }
+    elsif ($mode eq '>')  { $wordy_mode = 'writing';   }
+    elsif ($mode eq '>>') { $wordy_mode = 'appending'; }
+
+    return sprintf _FORMAT_OPEN, $file, $wordy_mode, $error if $wordy_mode;
+
+    Carp::confess("Internal autodie::exception error: Don't know how to format mode '$mode'.");
+
+}
 
 sub _format_open {
     my ($this) = @_;
 
     my @open_args = @{$this->args};
 
-    # We'll only handle 3 argument open for the moment.
-    if (@open_args != 3) {
+    # Use the default formatter for single-arg and many-arg open
+    if (@open_args <= 1 or @open_args >= 4) {
         return $this->format_default;
     }
 
+    # For two arg open, we have to extract the mode
+    if (@open_args == 2) {
+        my ($fh, $file) = @open_args;
+
+        if (ref($fh) eq "GLOB") {
+            $fh = '$fh';
+        }
+
+        my ($mode) = $file =~ m{
+            ^\s*                # Spaces before mode
+            (
+                (?>             # Non-backtracking subexp.
+                    <           # Reading
+                    |>>?        # Writing/appending
+                )
+            )
+            [^&]                # Not an ampersand (which means a dup)
+        }x;
+
+        # Have a funny mode?  Use the default format.
+        return $this->format_default if not defined $mode;
+
+        # Localising $! means perl make make it a pretty error for us.
+        local $! = $this->errno;
+
+        return $this->_format_open_with_mode($mode, $file, $!);
+    }
+
+    # Here we must be using three arg open.
+
     my $file = $open_args[2];
 
     local $! = $this->errno;
 
     my $mode = $open_args[1];
 
-    if    ($mode eq '<')  { return "Can't open '$file' for reading: '$!'"    }
-    elsif ($mode eq '>')  { return "Can't open '$file' for writing: '$!'"    }
-    elsif ($mode eq '>>') { return "Can't open '$file' for appending: '$!'"  }
+    local $@;
+
+    my $msg = eval { $this->_format_open_with_mode($mode, $file, $!); };
+
+    return $msg if $msg;
 
     # Default message (for pipes and odd things)
 
@@ -329,7 +459,7 @@ Primarily intended for use by format handlers.
 sub add_file_and_line {
     my ($this) = @_;
 
-    return sprintf(" at %s line %d", $this->file, $this->line);
+    return sprintf(" at %s line %d\n", $this->file, $this->line);
 }
 
 =head3 stringify
@@ -397,8 +527,23 @@ sub format_default {
     # Trim package name off dying sub for error messages.
     $call =~ s/.*:://;
 
-    return "Can't $call(".
-        join(q{, }, map { defined($_) ? "'$_'" : "undef" } @{$this->args()}) . "): $!" .
+    # Walk through all our arguments, and...
+    #
+    #   * Replace undef with the word 'undef'
+    #   * Replace globs with the string '$fh'
+    #   * Quote all other args.
+
+    my @args = @{ $this->args() };
+
+    foreach my $arg (@args) {
+       if    (not defined($arg))   { $arg = 'undef' }
+       elsif (ref($arg) eq "GLOB") { $arg = '$fh'   }
+       else                        { $arg = qq{'$arg'} }
+    }
+
+    # Format our beautiful error.
+
+    return "Can't $call(".  join(q{, }, @args) . "): $!" .
         $this->add_file_and_line;
 
     # TODO - Handle user-defined errors from hash.
@@ -412,6 +557,7 @@ sub format_default {
     my $error = autodie::exception->new(
         args => \@_,
         function => "CORE::open",
+        errno => $!,
     );
 
 
@@ -420,6 +566,11 @@ directly from an autodying function.  The C<function> argument
 is required, its the function we were trying to call that
 generated the exception.  The C<args> parameter is optional.
 
+The C<errno> value is optional.  In versions of C<autodie::exception>
+1.99 and earlier the code would try to automatically use the
+current value of C<$!>, but this was unreliable and is no longer
+supported.
+
 Atrributes such as package, file, and caller are determined
 automatically, and cannot be specified.
 
@@ -445,6 +596,9 @@ sub _init {
 
     my ($this, %args) = @_;
 
+    # Capturing errno here is not necessarily reliable.
+    my $original_errno = $!;
+
     our $init_called = 1;
 
     my $class = ref $this;
@@ -467,18 +621,42 @@ sub _init {
 
         next if $package->isa('Fatal');
         next if $package->isa($class);
+        next if $package->isa(__PACKAGE__);
         next if $file =~ /^\(eval\s\d+\)$/;
 
         last;
 
     }
 
+    # We now have everything correct, *except* for our subroutine
+    # name.  If it's __ANON__ or (eval), then we need to keep on
+    # digging deeper into our stack to find the real name.  However we
+    # don't update our other information, since that will be correct
+    # for our current exception.
+
+    my $first_guess_subroutine = $sub;
+
+    while (defined $sub and $sub =~ /^\(eval\)$|::__ANON__$/) {
+        $depth++;
+
+        $sub = (CORE::caller($depth))[3];
+    }
+
+    # If we end up falling out the bottom of our stack, then our
+    # __ANON__ guess is the best we can get.  This includes situations
+    # where we were called from the top level of a program.
+
+    if (not defined $sub) {
+        $sub = $first_guess_subroutine;
+    }
+
     $this->{$PACKAGE}{package} = $package;
     $this->{$PACKAGE}{file}    = $file;
     $this->{$PACKAGE}{line}    = $line;
     $this->{$PACKAGE}{caller}  = $sub;
     $this->{$PACKAGE}{package} = $package;
-    $this->{$PACKAGE}{errno}   = $!;
+
+    $this->{$PACKAGE}{errno}   = $args{errno} || 0;
 
     $this->{$PACKAGE}{args}    = $args{args} || [];
     $this->{$PACKAGE}{function}= $args{function} or
